// naumenn@stce.rwth-aachen.de

#pragma once

#include<vector>
#include<iostream>
#include<stdexcept>

template<typename T>
struct abstract_array {
  virtual ~abstract_array() {}
  virtual T& operator[](int)=0;
};

template<typename T>
class abstract_vector : public abstract_array<T> {
protected:
  T* v;
public:
  abstract_vector(int n) : v(new T[n]) {}
  ~abstract_vector() { delete [] v; }
};

template<typename T>
class abstract_std_vector : public abstract_array<T> {
protected:
  std::vector<T> *v;
public:
  abstract_std_vector(int n) : v(new std::vector<T>(n)) {}
   ~abstract_std_vector() { delete v; }
};

template<typename T>
class unsafe_vector : public abstract_vector<T> {
  using abstract_vector<T>::v;
public:
  unsafe_vector(int n) : abstract_vector<T>(n) { }
  T& operator[](int i) { return v[i]; }
};

template<typename T>
class safe_vector : public abstract_vector<T> {
  using abstract_vector<T>::v;
  int size;
public:
  safe_vector(int n) : abstract_vector<T>(n), size(n) { }
  T& operator[](int i) { 
    if (i<0||i>=size) throw std::out_of_range("i<0||i>=size");
    return v[i]; 
  }
};

template<typename T>
class unsafe_std_vector : public abstract_std_vector<T> {
  using abstract_std_vector<T>::v;
public:
  unsafe_std_vector(int n) : abstract_std_vector<T>(n) { }
  T& operator[](int i) { return (*v)[i]; }
};

template<typename T>
class safe_std_vector : public abstract_std_vector<T> {
  using abstract_std_vector<T>::v;
public:
  safe_std_vector(int n) : abstract_std_vector<T>(n) { }
  T& operator[](int i) { return v->at(i); }
};
